from anyio import create_task_group, sleep
from time import monotonic


class Monitor:
    def __init__(self, period: float = 0.01):
        """
        Create a Monitor with a [run()][anyioutils.Monitor.run] method that runs a task in the background and measures how saturated the event-loop is.
        This can also be used to detect (long) blocking calls in the event-loop.


        The Monitor can be used as an async context manager, in which case it will automatically run, or by launching its [run()][anyioutils.Monitor.run]
        method in the backgroup manually.

        Args:
            period: The period in seconds to make the measurement.
        """
        self._period = period
        self._result = 0
        self._iter = 1

    async def __aenter__(self) -> "Monitor":
        self._task_group = create_task_group()
        tg = await self._task_group.__aenter__()
        tg.start_soon(self.run)
        self._cancel_scope = tg.cancel_scope
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._cancel_scope.cancel()
        await self._task_group.__aexit__(None, None, None)

    async def run(self):
        """
        Run the Monitor. This has to be run in the background.
        """
        while True:
            t0 = monotonic()
            await sleep(self._period)
            t1 = monotonic()
            factor = (t1 - t0) / self._period
            self._result = self._result + (factor - self._result) / self._iter
            self._iter += 1

    @property
    def result(self) -> float:
        """
        The result of the measurement (greater than `1`). The closer to `1`, the less saturated the event-loop is. The greater, the more saturated the event-loop is.
        """
        return self._result



from __future__ import annotations

from contextlib import AsyncExitStack
from types import TracebackType
from typing import Any, Coroutine, TypeVar

from anyio import CancelScope, create_task_group

from ._task import Task, _task_group, create_task as _create_task

T = TypeVar("T")


class TaskGroup:
    """
    An [asynchronous context manager](https://docs.python.org/3/reference/datamodel.html#async-context-managers) holding a group of tasks. Tasks can be added to the group using [create_task()][anyioutils.TaskGroup.create_task]. All tasks are awaited when the context manager exits, unless they have been created with `background=True`.

    Tasks created down the call stack using [create_task()][anyioutils.create_task] or [start_task()][anyioutils.start_task] may not need to be passed a TaskGroup, since they could use this TaskGroup implicitly.
    """
    def __init__(self):
        self._background_tasks = []

    @property
    def cancel_scope(self) -> CancelScope:
        """
        Returns:
            The TaskGroup's [CancelScope](https://anyio.readthedocs.io/en/stable/api.html#anyio.CancelScope).
        """
        return self._task_group.cancel_scope

    async def __aenter__(self) -> "TaskGroup":
        async with AsyncExitStack() as exit_stack:
            self._task_group = await exit_stack.enter_async_context(create_task_group())
            self._token = _task_group.set(self._task_group)
            self._exit_stack = exit_stack.pop_all()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        _task_group.reset(self._token)
        for task in self._background_tasks:
            task.cancel()
        await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    def create_task(self, coro: Coroutine[Any, Any, T], *, name: str | None = None, background: bool = False) -> Task[T]:
        """
        Create a task in this task group.

        Args:
            background: Whether to run the task in the background, in which case it will be cancelled when the task group exits.
            name: The task name.
        Returns:
            The created Task.
        """
        task = _create_task(coro, self._task_group, name=name)
        if background:
            self._background_tasks.append(task)
        return task



from anyio import Event as _Event


class Event:
    """
    An event object. Not thread-safe.

    An event can be used to notify multiple tasks that some event has happened.

    An Event object manages an internal flag that can be set to `True` with the [set()][anyioutils.Event.set] method and reset to `False` with the [clear()][anyioutils.Event.clear] method. The [wait()][anyioutils.Event.wait] method blocks until the flag is set to `True`. The flag is set to `False` initially.
    """
    def __init__(self) -> None:
        self._event = _Event()

    async def wait(self) -> bool:
        """
        Wait until the event is set.

        Returns:
            If the event is set, return `True` immediately. Otherwise block until another task calls [set()][anyioutils.Event.set].
        """
        await self._event.wait()
        return True

    def set(self) -> None:
        """
        Set the event.

        All tasks waiting for event to be set will be immediately awakened.
        """
        self._event.set()

    def is_set(self) -> bool:
        """
        Returns:
            `True` if the event is set.
        """
        return self._event.is_set()

    def clear(self) -> None:
        """
        Clear (unset) the event.

        Tasks awaiting on [wait()][anyioutils.Event.wait] will now block until the [set()][anyioutils.Event.set] method is called again.
        """
        if self._event.is_set():
            self._event = _Event()



class CancelledError(BaseException):
    """
    The operation has been cancelled.
    """


class InvalidStateError(BaseException):
    """
    Invalid internal state of [Task][anyioutils.Task] or [Future][anyioutils.Future].

    Can be raised in situations like setting a result value for a Future object that already has a result value set.
    """



from typing import Generic, TypeVar

from anyio import create_memory_object_stream


T = TypeVar("T")


class Queue(Generic[T]):
    def __init__(self, maxsize: int = 0):
        """
        A first in, first out (FIFO) queue.

        If *maxsize* is less than or equal to zero, the queue size is infinite. If it is an integer greater than `0`, then `await put()` blocks when the queue reaches *maxsize* until an item is removed by [get()][anyioutils.Queue.get].

        """
        self._maxsize = maxsize
        max_buffer_size = float("inf") if maxsize <= 0 else maxsize
        self._send_stream, self._receive_stream = create_memory_object_stream[T](max_buffer_size=max_buffer_size)

    @property
    def maxsize(self) -> int:
        """
        Number of items allowed in the queue.
        """
        return self._maxsize

    def qsize(self) -> int:
        """
        Returns:
            The number of items in the queue.
        """
        return self._send_stream.statistics().current_buffer_used

    def empty(self) -> bool:
        """
        Returns:
            `True` if the queue is empty, `False` otherwise.
        """
        return self._send_stream.statistics().current_buffer_used == 0

    def full(self) -> bool:
        """
        Returns:
            `True` if there are [maxsize][anyioutils.Queue.maxsize] items in the queue.
        """
        statistics = self._send_stream.statistics()
        return statistics.current_buffer_used == statistics.max_buffer_size

    async def put(self, item: T) -> None:
        """
        Put an item into the queue. If the queue is full, wait until a free slot is available before adding the item.

        Args:
            item: The item to put into the queue.
        """
        await self._send_stream.send(item)

    def put_nowait(self, item: T) -> None:
        """
        Put an item into the queue without blocking.

        Args:
            item: The item to put into the queue.
        """
        self._send_stream.send_nowait(item)

    async def get(self) -> T:
        """
        Remove and return an item from the queue. If queue is empty, wait until an item is available.

        Returns:
            The item from the queue.
        """
        return await self._receive_stream.receive()

    def get_nowait(self) -> T:
        """
        Returns:
            An item if one is immediately available.
        """
        return self._receive_stream.receive_nowait()

    def __del__(self) -> None:
        self._send_stream.close()
        self._receive_stream.close()



from __future__ import annotations

from typing import Any, Iterable, Literal

from anyio import create_memory_object_stream, move_on_after
from anyio.abc import TaskGroup
from anyio.streams.memory import MemoryObjectSendStream

from ._future import Future
from ._task import Task, _task_group


ALL_COMPLETED: Literal["ALL_COMPLETED"] = "ALL_COMPLETED"
FIRST_COMPLETED: Literal["FIRST_COMPLETED"] = "FIRST_COMPLETED"
FIRST_EXCEPTION: Literal["FIRST_EXCEPTION"] = "FIRST_EXCEPTION"


async def _run_and_put_task(
    task: Task | Future,
    send_stream: MemoryObjectSendStream[Any],
):
    exc = None
    try:
        await task.wait()
    except Exception as e:
        exc = e
    try:
        await send_stream.send((task, exc))
    except Exception:
        pass


async def wait(
    aws: Iterable[Task | Future],
    task_group: TaskGroup | None = None,
    *,
    timeout: float | int | None = None,
    return_when: Literal["ALL_COMPLETED", "FIRST_COMPLETED", "FIRST_EXCEPTION"] = ALL_COMPLETED,
) -> tuple[set[Task | Future], set[Task | Future]]:
    """
    Run [Future][anyioutils.Future] and [Task][anyioutils.Task] instances in the *aws* iterable concurrently and block until the condition specified by `return_when`.

    *return_when* indicates when this function should return. It must be one of the following constants:

    | Constant | Description |
    | --- | --- |
    | anyioutils.FIRST_COMPLETED | The function will return when any future finishes or is cancelled. |
    | anyioutils.FIRST_EXCEPTION | The function will return when any future finishes by raising an exception. If no future raises an exception then it is equivalent to [`ALL_COMPLETED`. |
    | anyioutils.ALL_COMPLETED | The function will return when all futures finish or are cancelled. |


    Args:
        timeout: If specified, can be used to control the maximum number of seconds to wait before returning. Note that this function does not raise [TimeoutError](https://docs.python.org/3/library/exceptions.html#TimeoutError). Futures or Tasks that aren't done when the timeout occurs are simply returned in the second set.
        return_when: Indicates when this function should return. It must be one of the following constants:

    Returns:
        Two sets of Tasks/Futures: `(done, pending)`.
    """
    if task_group is None:
        task_group = _task_group.get()
    for aw in aws:
        if not isinstance(aw, (Task, Future)):
            raise TypeError(f"Pass tasks or futures, not {type(aw)}")
    if timeout is None:
        timeout = float("inf")
    done = set()
    pending = set(aws)
    send_stream, receive_stream = create_memory_object_stream[Any]()
    async with send_stream, receive_stream:
        for task in aws:
            task_group.start_soon(_run_and_put_task, task, send_stream)
        with move_on_after(timeout):
            async for aw_exc in receive_stream:
                aw, exc = aw_exc
                done.add(aw)
                pending.remove(aw)
                if return_when == FIRST_EXCEPTION and exc is not None:
                    break
                if return_when == FIRST_COMPLETED:
                    break
                if not pending:
                    break
        return done, pending



from __future__ import annotations

from inspect import isawaitable
from sys import version_info
from collections.abc import Awaitable, Coroutine
from contextvars import ContextVar
from typing import Any, Callable, Generic, TypeVar

from anyio import Event, create_task_group
from anyio.abc import TaskGroup

from ._exceptions import CancelledError, InvalidStateError
from ._queue import Queue

if version_info < (3, 10):  # pragma: no cover
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias  # type: ignore[attr-defined,no-redef]

if version_info < (3, 11):  # pragma: no cover
    from exceptiongroup import BaseExceptionGroup  # type: ignore[import-not-found]


T = TypeVar("T")
ExceptionHandler: TypeAlias = Callable[[BaseException], Any]
_task_group: ContextVar[TaskGroup] = ContextVar("_task_group")


async def ensure_awaitable(obj: T | Awaitable[T]) -> T:
    if isawaitable(obj):
        return await obj
    return obj


class Task(Generic[T]):
    """
    A Future-like object that runs a Python [coroutine](https://docs.python.org/3/library/asyncio-task.html#coroutine). Not thread-safe.

    Tasks are used to run coroutines in event loops. If a coroutine awaits on a Future, the Task suspends the execution of the coroutine and waits for the completion of the Future. When the Future is done, the execution of the wrapped coroutine resumes.

    Event loops use cooperative scheduling: an event loop runs one Task at a time. While a Task awaits for the completion of a Future, the event loop runs other Tasks, callbacks, or performs IO operations.

    Use the high-level [anyioutils.create_task()][anyioutils.create_task] function to create Tasks. Manual instantiation of Tasks is discouraged.

    To cancel a running Task use the [cancel()][anyioutils.Task.cancel] method. Calling it will cause the Task to throw a [CancelledError][anyioutils.CancelledError] exception into the wrapped coroutine. If a coroutine is awaiting on a Future object during cancellation, the Future object will be cancelled.

    [cancelled()][anyioutils.Task.cancelled] can be used to check if the Task was cancelled.
    """
    _done_callbacks: list[Callable[[Task], None]]
    _exception: BaseException | None
    _result: T | None

    def __init__(self, coro: Coroutine[Any, Any, T], exception_handler: ExceptionHandler | None = None) -> None:
        self._coro = coro
        self._exception_handler = exception_handler
        self._has_result = False
        self._has_exception = False
        self._coro_started = False
        self._cancelled_event = Event()
        self._raise_cancelled_error = True
        self._done_callbacks = []
        self._done_event = Event()
        self._exception = None
        self._waiting = False
        self._started_value = Queue[Any]()

    def _call_callbacks(self) -> None:
        exceptions = []
        for callback in self._done_callbacks:
            try:
                callback(self)
            except BaseException as exc:
                exceptions.append(exc)
        if not exceptions:
            return
        if len(exceptions) == 1:
            raise exceptions[0]
        raise BaseExceptionGroup("Error while calling callbacks", exceptions)

    async def _wait_result(self, task_group: TaskGroup) -> None:
        try:
            self._coro_started = True
            self._result = await self._coro
            self._has_result = True
        except BaseException as exc:
            if self._exception_handler is None or not await ensure_awaitable(self._exception_handler(exc)):
                self._exception = exc
                self._has_exception = True
            else:
                self._result = None
                self._has_result = True
        self._done_event.set()
        task_group.cancel_scope.cancel()
        self._call_callbacks()

    async def _wait_cancelled(self, task_group: TaskGroup) -> None:
        await self._cancelled_event.wait()
        task_group.cancel_scope.cancel()

    def cancel(self, raise_exception: bool = False) -> None:
        """
        Request the Task to be cancelled.
        """
        if not self._coro_started:
            self._coro.close()
        self._done_event.set()
        self._cancelled_event.set()
        self._raise_cancelled_error = raise_exception
        self._call_callbacks()

    def cancelled(self) -> bool:
        """
        Returns:
            `True` if the Task is *cancelled*.
        """
        return self._cancelled_event.is_set()

    async def wait(self) -> T | None:
        """
        Wait for the Task to be *done* or *cancelled*.

        Returns:
            The return value of the coroutine, if not *cancelled*, otherwise `None`.
        """
        if self._waiting:
            await self._done_event.wait()
        self._waiting = True
        if self._has_result:
            return self._result
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError
            return None
        if self._has_exception:
            assert self._exception is not None
            raise self._exception

        async with create_task_group() as tg:
            tg.start_soon(self._wait_result, tg)
            tg.start_soon(self._wait_cancelled, tg)

        if self._has_result:
            return self._result
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError
            return None
        if self._has_exception:
            assert self._exception is not None
            raise self._exception

        return None  # pragma: nocover

    def done(self) -> bool:
        """
        A Task is *done* when the wrapped coroutine either returned a value, raised an exception, or the Task was cancelled.

        Return:
            `True` if the Task is done.
        """
        return self._done_event.is_set()

    def result(self) -> T | None:
        """
        If the Task is *done*, the result of the wrapped coroutine is returned (or if the coroutine raised an exception, that exception is re-raised).

        If the Task has been *cancelled*, this method raises a [CancelledError][anyioutils.CancelledError] exception.

        If the Task's result is't yet available, this method raises an [InvalidStateError][anyioutils.InvalidStateError] exception.

        Return:
            The result of the Task.
        """
        if self._cancelled_event.is_set():
            raise CancelledError
        if self._has_result:
            return self._result
        if self._has_exception:
            assert self._exception is not None
            raise self._exception
        raise InvalidStateError

    def exception(self) -> BaseException | None:
        """
        If the wrapped coroutine raised an exception that exception is returned. If the wrapped coroutine returned normally this method returns `None`.

        If the Task has been *cancelled*, this method raises a [CancelledError][anyioutils.CancelledError] exception.

        If the Task isn't done yet, this method raises an [InvalidStateError][anyioutils.InvalidStateError] exception.

        Returns:
            The exception of the Task.
        """
        if not self._done_event.is_set():
            raise InvalidStateError
        if self._cancelled_event.is_set():
            raise CancelledError
        return self._exception

    def add_done_callback(self, callback: Callable[[Task], None]) -> None:
        """
        Add a callback to be run when the Task is *done*.

        This method should only be used in low-level callback-based code.

        See the documentation of [Future.add_done_callback()][anyioutils.Future.add_done_callback] for more details.
        """
        self._done_callbacks.append(callback)
        if self._done_event.is_set():
            callback(self)

    def remove_done_callback(self, callback: Callable[[Task], None]) -> int:
        """
        Remove *callback* from the callbacks list.

        This method should only be used in low-level callback-based code.

        See the documentation of [Future.remove_done_callback()][anyioutils.Future.remove_done_callback] for more details.
        """
        count = self._done_callbacks.count(callback)
        for _ in range(count):
            self._done_callbacks.remove(callback)
        return count

    async def wait_started(self) -> Any:
        """
        Wait for the task to be [started](https://anyio.readthedocs.io/en/stable/tasks.html#starting-and-initializing-tasks).
        The task must have been created with [start_task()][anyioutils.start_task], not [create_task][anyioutils.create_task].

        Returns:
            The started value.
        """
        return await self._started_value.get()


def create_task(
    coro: Coroutine[Any, Any, T],
    task_group: TaskGroup | None = None,
    *,
    name: str | None = None,
    exception_handler: ExceptionHandler | None = None,
) -> Task[T]:
    """
    Wrap the *coro* [coroutine](https://docs.python.org/3/library/asyncio-task.html#coroutine) into a [Task][anyioutils.Task] and schedule its execution.

    Args:
        task_group: An optional [TaskGroup](https://anyio.readthedocs.io/en/stable/api.html#anyio.abc.TaskGroup) (from AnyIO) to run the Task in. If not provided, a [TaskGroup][anyioutils.TaskGroup] (from `anyioutils`) will be looked up the call stack and used if found.
        exception_handler: An optional exception handler. When an exception occurs in the Task, the exception handler is called with the exception. The exception is considered to be handled if the exception handler returns `True`, otherwise the exception is raised.

    Returns:
        The Task object.
    """
    task = Task[T](coro, exception_handler)
    if task_group is None:
        task_group = _task_group.get()
    task_group.start_soon(task.wait, name=name)
    return task


def start_task(
    async_fn: Callable[..., Awaitable[Any]],
    task_group: TaskGroup | None = None,
    *,
    name: str | None = None,
    exception_handler: ExceptionHandler | None = None,
) -> Task[None]:
    """
    Create a [coroutine](https://docs.python.org/3/library/asyncio-task.html#coroutine) from the *async_fn* async function, wrap it into a [Task][anyioutils.Task] and schedule its execution.

    The Task's [wait()][anyioutils.Task.wait] method will only return `None`, but its [wait_started()][anyioutils.Task.wait_started] method will return its started value.

    Args:
        task_group: An optional [TaskGroup](https://anyio.readthedocs.io/en/stable/api.html#anyio.abc.TaskGroup) (from AnyIO) to run the Task in. If not provided, a [TaskGroup][anyioutils.TaskGroup] (from `anyioutils`) will be looked up the call stack and used, if found.
        exception_handler: An optional exception handler. When an exception occurs in the Task, the exception handler is called with the exception. The exception is considered to be handled if the exception handler returns `True`, otherwise the exception is raised.

    Returns:
        The Task object.
    """
    async_function_wrapper = AsyncFunctionWrapper(async_fn)
    task = Task[None](async_function_wrapper.get_coro(), exception_handler)
    async_function_wrapper.set_task(task)
    if task_group is None:
        task_group = _task_group.get()
    task_group.start_soon(task.wait, name=name)
    return task


class AsyncFunctionWrapper:
    def __init__(self, async_fn: Callable[..., Awaitable[T]]) -> None:
        self._async_fn = async_fn

    def set_task(self, task: Task) -> None:
        self._task = task

    async def get_coro(self) -> None:
        async with create_task_group() as tg:
            started_value = await tg.start(self._async_fn)
            await self._task._started_value.put(started_value)



from __future__ import annotations

from sys import version_info
from typing import Callable, Generic, TypeVar

from anyio import Event, create_task_group
from anyio.abc import TaskGroup

from ._exceptions import CancelledError, InvalidStateError

if version_info < (3, 11):  # pragma: no cover
    from exceptiongroup import BaseExceptionGroup  # type: ignore[import-not-found]

T = TypeVar("T")


class Future(Generic[T]):
    """
    A Future represents an eventual result of an asynchronous operation. Not thread-safe.

    Future is an [awaitable](https://docs.python.org/3/glossary.html#term-awaitable) object.
    Coroutines can await on Future objects until they either have a result or an exception set,
    or until they are cancelled. A Future can be awaited multiple times and the result is same.

    Typically Futures are used to enable low-level callback-based code to interoperate with high-level async/await code.
    """

    _done_callbacks: list[Callable[[Future], None]]
    _exception: BaseException | None

    def __init__(self) -> None:
        self._result_event = Event()
        self._exception_event = Event()
        self._cancelled_event = Event()
        self._raise_cancelled_error = True
        self._done_callbacks = []
        self._done_event = Event()
        self._exception = None
        self._waiting = False

    def _call_callbacks(self) -> None:
        exceptions = []
        for callback in self._done_callbacks:
            try:
                callback(self)
            except BaseException as exc:
                exceptions.append(exc)
        if not exceptions:
            return
        if len(exceptions) == 1:
            raise exceptions[0]
        raise BaseExceptionGroup("Error while calling callbacks", exceptions)

    async def _wait_result(self, task_group: TaskGroup) -> None:
        await self._result_event.wait()
        task_group.cancel_scope.cancel()

    async def _wait_exception(self, task_group: TaskGroup) -> None:
        await self._exception_event.wait()
        task_group.cancel_scope.cancel()

    async def _wait_cancelled(self, task_group: TaskGroup) -> None:
        await self._cancelled_event.wait()
        task_group.cancel_scope.cancel()

    def cancel(self, raise_exception: bool = False) -> bool:
        """
        Cancel the Future and schedule callbacks.

        If the Future is already *done* or *cancelled*, return `False`.
        Otherwise, change the Future's state to *cancelled*, schedule the callbacks, and return `True`.

        Args:
            raise_exception: Whether to raise a [CancelledError][anyioutils.CancelledError].

        Returns:
            `False` if the Future is already *done* or *cancelled*, `True` otherwise.
        """
        if self._done_event.is_set() or self._cancelled_event.is_set():
            return False

        self._done_event.set()
        self._cancelled_event.set()
        self._raise_cancelled_error = raise_exception
        self._call_callbacks()
        return True

    def cancelled(self) -> bool:
        """
        Returns:
            `True` if the Future was cancelled, `False` otherwise.
        """
        return self._cancelled_event.is_set()

    async def wait(self) -> T | None:
        """
        Wait for the Future to be *done* or *cancelled*.

        Returns:
            The Furure's return value.
        """
        if self._waiting:
            await self._done_event.wait()
        self._waiting = True
        if self._result_event.is_set():
            return self._result
        if self._exception_event.is_set():
            assert self._exception is not None
            raise self._exception
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError

        async with create_task_group() as tg:
            tg.start_soon(self._wait_result, tg)
            tg.start_soon(self._wait_exception, tg)
            tg.start_soon(self._wait_cancelled, tg)

        if self._result_event.is_set():
            return self._result
        if self._exception_event.is_set():
            assert self._exception is not None
            raise self._exception
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError

        return None  # pragma: nocover

    def done(self) -> bool:
        """
        A Future is *done* if it was *cancelled* or if it has a result or an exception set with [set_result()][anyioutils.Future.set_result] or [set_exception()][anyioutils.Future.set_exception] calls.

        Returns:
            `True` if the Future is *done*.
        """
        return self._done_event.is_set()

    def set_result(self, value: T) -> None:
        """
        Mark the Future as *done* and set its result.

        Raises:
            InvalidStateError: The Future is already *done*.
        """
        if self._done_event.is_set():
            raise InvalidStateError
        self._done_event.set()
        self._result = value
        self._result_event.set()
        self._call_callbacks()

    def result(self) -> T:
        """
        If the Future is *done* and has a result set by the [set_result()][anyioutils.Future.set_result] method, the result value is returned.

        If the Future is *done* and has an exception set by the [set_exception()][anyioutils.Future.set_exception] method, this method raises the exception.

        If the Future has been *cancelled*, this method raises a [CancelledError][anyioutils.CancelledError] exception.

        If the Future’s result isn’t yet available, this method raises an [InvalidStateError][anyioutils.InvalidStateError] exception.

        Returns:
            The result of the Future.
        """
        if self._cancelled_event.is_set():
            raise CancelledError
        if self._result_event.is_set():
            return self._result
        if self._exception_event.is_set():
            assert self._exception is not None
            raise self._exception
        raise InvalidStateError

    def set_exception(self, value: BaseException) -> None:
        """
        Mark the Future as *done* and set an exception.

        Raises:
            InvalidStateError: The Future is already *done*.
        """
        if self._done_event.is_set():
            raise InvalidStateError
        self._done_event.set()
        self._exception = value
        self._exception_event.set()
        self._call_callbacks()

    def exception(self) -> BaseException | None:
        """
        The exception (or `None` if no exception was set) is returned only if the Future is *done*.

        If the Future has been *cancelled*, this method raises a [CancelledError][anyioutils.CancelledError] exception.

        If the Future isn’t *done* yet, this method raises an [InvalidStateError][anyioutils.InvalidStateError] exception.

        Returns:
            The exception that was set on this Future.
        """
        if not self._done_event.is_set():
            raise InvalidStateError
        if self._cancelled_event.is_set():
            raise CancelledError
        return self._exception

    def add_done_callback(self, callback: Callable[[Future], None]) -> None:
        """
        Add a callback to be run when the Future is *done*.

        The *callback* is called with the Future object as its only argument.

        If the Future is already *done* when this method is called, the callback is scheduled immediately.
        """
        self._done_callbacks.append(callback)
        if self._done_event.is_set():
            callback(self)

    def remove_done_callback(self, callback: Callable[[Future], None]) -> int:
        """
        Remove *callback* from the callbacks list.

        Returns:
            The number of callbacks removed, which is typically 1, unless a callback was added more than once.
        """
        count = self._done_callbacks.count(callback)
        for _ in range(count):
            self._done_callbacks.remove(callback)
        return count
